from dkg.managers.async_manager import AsyncRequestManager
from dkg.method import Method
from dkg.constants import OperationStatuses, ErrorType, Status
from dkg.utils.node_request import NodeRequest
from dkg.modules.async_module import AsyncModule
from typing import Dict, Any
from dkg.types import UAL
from dkg.dataclasses import NodeResponseDict
import asyncio


class AsyncNodeService(AsyncModule):
    def __init__(self, manager: AsyncRequestManager):
        self.manager = manager

    _info = Method(NodeRequest.info)
    _get_operation_result = Method(NodeRequest.get_operation_result)
    _finality_status = Method(NodeRequest.finality_status)
    _ask = Method(NodeRequest.ask)
    _get_bid_suggestion = Method(NodeRequest.bid_suggestion)
    _publish = Method(NodeRequest.publish)
    _get = Method(NodeRequest.get)
    _query = Method(NodeRequest.query)

    async def info(self) -> NodeResponseDict:
        return await self._info()

    async def get_operation_result(
        self,
        operation_id: str,
        operation: str,
        max_retries: int,
        frequency: int,
    ) -> Dict[str, Any]:
        response = {"status": OperationStatuses.PENDING.value}

        retries = 0

        while True:
            if retries > max_retries:
                response["data"] = {
                    "errorType": ErrorType.DKG_CLIENT_ERROR.value,
                    "errorMessage": "Unable to get results. Max number of retries reached.",
                }
                break

            if retries > 0:
                await asyncio.sleep(frequency)

            retries += 1

            try:
                result = await self._get_operation_result(
                    operation=operation,
                    operation_id=operation_id,
                )
                response = {"data": result}
            except Exception:
                response = {"data": {"status": Status.NETWORK_ERROR.value}}

            # Check completion conditions
            if (
                response["data"].get("status") == OperationStatuses.COMPLETED.value
                or response["data"].get("status") == OperationStatuses.FAILED.value
            ):
                break

        return response["data"]

    async def finality_status(
        self,
        ual: UAL,
        required_confirmations: int,
        max_number_of_retries: int,
        frequency: int,
    ):
        retries = 0
        finality = 0

        while finality < required_confirmations and retries <= max_number_of_retries:
            if retries > max_number_of_retries:
                raise Exception(
                    f"Unable to achieve required confirmations. "
                    f"Max number of retries ({max_number_of_retries}) reached."
                )

            # Sleep between attempts (except for first try)
            if retries > 0:
                await asyncio.sleep(frequency)

            retries += 1

            try:
                response = await self._finality_status(ual)
                finality = response.get("finality", 0)
            except Exception:
                finality = 0

        return finality

    async def ask(
        self,
        ual: UAL,
        required_confirmations: int,
        max_number_of_retries: int,
        frequency: int,
    ):
        finality_id = 0
        retries = 0

        while finality_id < required_confirmations and retries < max_number_of_retries:
            if retries > max_number_of_retries:
                raise Exception(
                    f"Unable to achieve required confirmations. "
                    f"Max number of retries ({max_number_of_retries}) reached."
                )

            if retries > 0:
                await asyncio.sleep(frequency)

            retries += 1

            try:
                try:
                    response = await self._ask(
                        ual=ual, minimumNumberOfNodeReplications=required_confirmations
                    )
                except Exception as e:
                    response = None
                    print(f"failed: {e}")

                if response is not None:
                    operation_id = response.json().get("operationId", 0)
                    if operation_id >= required_confirmations:
                        finality_id = operation_id

            except Exception as e:
                finality_id = 0
                print(f"Retry {retries + 1}/{max_number_of_retries} failed: {e}")

            return finality_id

    async def publish(
        self,
        dataset_root,
        dataset,
        blockchain_id,
        hash_function_id,
        minimum_number_of_node_replications,
    ):
        return await self._publish(
            dataset_root,
            dataset,
            blockchain_id,
            hash_function_id,
            minimum_number_of_node_replications,
        )

    async def get(
        self,
        ual_with_state,
        content_type,
        include_metadata,
        hash_function_id,
        paranet_ual,
        subject_ual,
    ):
        return await self._get(
            ual_with_state,
            content_type,
            include_metadata,
            hash_function_id,
            paranet_ual,
            subject_ual,
        )

    async def query(
        self,
        query,
        query_type,
        repository,
        paranet_ual,
    ):
        return await self._query(query, query_type, repository, paranet_ual)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any

import requests
from dkg.dataclasses import HTTPRequestMethod, NodeResponseDict
from dkg.exceptions import HTTPRequestMethodNotSupported, NodeRequestError
from dkg.types import URI
from requests.exceptions import HTTPError, ConnectionError, Timeout, RequestException
from dkg.providers.node.base_node_http import BaseNodeHTTPProvider


class NodeHTTPProvider(BaseNodeHTTPProvider):
    def __init__(
        self,
        endpoint_uri: URI | str,
        api_version: str = "v1",
        auth_token: str | None = None,
    ):
        super().__init__(endpoint_uri, api_version, auth_token)

    def make_request(
        self,
        method: HTTPRequestMethod,
        path: str,
        params: dict[str, Any] = {},
        data: dict[str, Any] = {},
    ) -> NodeResponseDict:
        url = f"{self.url}/{path}"

        try:
            if method == HTTPRequestMethod.GET:
                response = requests.get(url, params=params, headers=self.headers)
            elif method == HTTPRequestMethod.POST:
                response = requests.post(url, json=data, headers=self.headers)
            else:
                raise HTTPRequestMethodNotSupported(
                    f"{method.name} method isn't supported"
                )

            response.raise_for_status()

            try:
                return NodeResponseDict(response.json())
            except ValueError as err:
                raise NodeRequestError(f"JSON decoding failed: {err}")

        except (HTTPError, ConnectionError, Timeout, RequestException) as err:
            raise NodeRequestError(f"Request failed: {err}")



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import copy
import hashlib
from typing import Callable

from dkg.exceptions import LeafNotInTree
from dkg.types import HexStr
from eth_abi.packed import encode_packed
from hexbytes import HexBytes
from web3 import Web3


def solidity_keccak256(data: HexStr) -> HexStr:
    bytes_hash: HexBytes = Web3.solidity_keccak(
        ["bytes"],
        [data],
    )

    return bytes_hash.to_0x_hex()


def hash_assertion_with_indexes(
    leaves: list[str],
    hash_function: str | Callable[[str], HexStr] = solidity_keccak256,
    sort: bool = True,
) -> list[HexStr]:
    if sort:
        leaves.sort()

    return list(
        map(
            hash_function,
            [
                encode_packed(
                    ["bytes32", "uint256"],
                    [Web3.solidity_keccak(["string"], [leaf]), i],
                )
                for i, leaf in enumerate(leaves)
            ],
        )
    )


class MerkleTree:
    def __init__(
        self,
        leaves: list[str],
        hash_function: str | Callable[[str], HexStr] = solidity_keccak256,
        sort_leaves: bool = False,
        sort_pairs: bool = False,
    ):
        self.hash_function = self._set_hash_function(hash_function)
        self.sort_leaves = sort_leaves
        self.sort_pairs = sort_pairs
        self.leaves = self._process_leaves(leaves)
        self.tree = self.build_tree()

    @property
    def root(self) -> HexStr:
        return self.tree[0][0]

    def build_tree(self) -> list[list[HexStr]]:
        tree = [self.leaves]

        while len(level := tree[-1]) > 1:
            next_level = []
            for h1, h2 in zip(level[::2], level[1::2] + [None]):
                if h2:
                    next_level.append(
                        self.hash_function(
                            h1 + h2[2:]
                            if not self.sort_pairs
                            else "0x" + "".join(sorted([h1[2:], h2[2:]]))
                        )
                    )
                else:
                    next_level.append(h1)

            tree.append(next_level)

        tree.reverse()
        return tree

    def proof(self, leaf: HexStr, index: int | None = None) -> list[HexStr]:
        if index is None:
            for i, t_leaf in enumerate(self.leaves):
                if leaf == t_leaf:
                    index = i
                    break

        if index is None:
            raise LeafNotInTree(f"{leaf} is not a part of the Merkle Tree.")

        proof = []
        levels = copy.deepcopy(self.tree[:0:-1])
        for level in levels:
            if (len(level) % 2) == 1:
                level.append(level[-1])

            if (index % 2) == 1:
                proof.append(level[index - 1])
            else:
                proof.append(level[index + 1])

            index //= 2

        return proof

    def verify(self, proof: list[HexStr], leaf: HexStr) -> bool:
        if self.sort_pairs:
            hash = leaf
            for p in proof:
                if hash == p:
                    continue

                hash = self.hash_function("".join(sorted([hash, p])))

        else:
            for i, t_leaf in enumerate(self.leaves):
                if leaf == t_leaf:
                    index = i
                    break

            if index is None:
                raise LeafNotInTree(f"{leaf} is not a part of the Merkle Tree.")

            hash = leaf
            for p in proof:
                if hash == p:
                    continue

                is_left = (index % 2) == 0
                hash = self.hash_function("".join([hash, p] if is_left else [p, hash]))
                index //= 2

        return hash == self.root

    def _process_leaves(self, leaves: list[str | HexStr]) -> list[HexStr]:
        if self.sort_leaves:
            leaves.sort()

        return leaves

    def _set_hash_function(
        self, hash_function: str | Callable[[str], HexStr]
    ) -> Callable[[str], HexStr]:
        if (
            isinstance(hash_function, str)
            and hash_function in hashlib.algorithms_available
        ):
            return lambda data: getattr(hashlib, hash_function)(
                data.encode()
            ).hexdigest()
        elif isinstance(hash_function, Callable):
            return hash_function
        else:
            raise ValueError()



from dkg.modules.async_module import AsyncModule
from dkg.managers.async_manager import AsyncRequestManager
from dkg.utils.blockchain_request import BlockchainRequest
from dkg.method import Method
from dkg.constants import ZERO_ADDRESS
from web3 import Web3
from typing import Optional
from dkg.types import Address, UAL
from dkg.utils.blockchain_request import KnowledgeCollectionResult, AllowanceResult
from dkg.utils.ual import parse_ual


class AsyncBlockchainService(AsyncModule):
    def __init__(self, manager: AsyncRequestManager):
        self.manager = manager

    _owner = Method(BlockchainRequest.owner_of)
    _get_contract_address = Method(BlockchainRequest.get_contract_address)
    _get_current_allowance = Method(BlockchainRequest.allowance)
    _increase_allowance = Method(BlockchainRequest.increase_allowance)
    _decrease_allowance = Method(BlockchainRequest.decrease_allowance)
    _create_knowledge_collection = Method(BlockchainRequest.create_knowledge_collection)
    _mint_knowledge_collection = Method(BlockchainRequest.mint_knowledge_collection)
    _get_asset_storage_address = Method(BlockchainRequest.get_asset_storage_address)
    _key_is_operational_wallet = Method(BlockchainRequest.key_is_operational_wallet)
    _time_until_next_epoch = Method(BlockchainRequest.time_until_next_epoch)
    _epoch_length = Method(BlockchainRequest.epoch_length)
    _get_stake_weighted_average_ask = Method(
        BlockchainRequest.get_stake_weighted_average_ask
    )
    _get_block = Method(BlockchainRequest.get_block)

    async def decrease_knowledge_collection_allowance(
        self,
        allowance_gap: int,
    ):
        knowledge_collection_address = await self._get_contract_address(
            "KnowledgeCollection"
        )
        await self._decrease_allowance(knowledge_collection_address, allowance_gap)

    async def increase_knowledge_collection_allowance(
        self,
        sender: str,
        token_amount: str,
    ) -> AllowanceResult:
        """
        Increases the allowance for knowledge collection if necessary.

        Args:
            sender: The address of the sender
            token_amount: The amount of tokens to check/increase allowance for

        Returns:
            AllowanceResult containing whether allowance was increased and the gap
        """
        knowledge_collection_address = await self._get_contract_address(
            "KnowledgeCollection"
        )

        allowance = await self._get_current_allowance(
            sender, knowledge_collection_address
        )
        allowance_gap = int(token_amount) - int(allowance)

        if allowance_gap > 0:
            await self._increase_allowance(knowledge_collection_address, allowance_gap)

            return AllowanceResult(
                allowance_increased=True, allowance_gap=allowance_gap
            )

        return AllowanceResult(allowance_increased=False, allowance_gap=allowance_gap)

    async def create_knowledge_collection(
        self,
        request: dict,
        paranet_ka_contract: Optional[Address] = None,
        paranet_token_id: Optional[int] = None,
    ) -> KnowledgeCollectionResult:
        """
        Creates a knowledge collection on the blockchain.

        Args:
            request: dict containing all collection parameters
            paranet_ka_contract: Optional paranet contract address
            paranet_token_id: Optional paranet token ID
            blockchain: Blockchain configuration

        Returns:
            KnowledgeCollectionResult containing collection ID and transaction receipt

        Raises:
            BlockchainError: If the collection creation fails
        """
        sender = self.manager.blockchain_provider.account.address
        allowance_increased = False
        allowance_gap = 0

        try:
            # Handle allowance
            if request.get("paymaster") and request.get("paymaster") != ZERO_ADDRESS:
                pass
            else:
                allowance_result = await self.increase_knowledge_collection_allowance(
                    sender=sender,
                    token_amount=request.get("tokenAmount"),
                )
                allowance_increased = allowance_result.allowance_increased
                allowance_gap = allowance_result.allowance_gap

            if not paranet_ka_contract and not paranet_token_id:
                receipt = await self._create_knowledge_collection(
                    request.get("publishOperationId"),
                    Web3.to_bytes(hexstr=request.get("merkleRoot")),
                    request.get("knowledgeAssetsAmount"),
                    request.get("byteSize"),
                    request.get("epochs"),
                    request.get("tokenAmount"),
                    request.get("isImmutable"),
                    request.get("paymaster"),
                    request.get("publisherNodeIdentityId"),
                    Web3.to_bytes(hexstr=request.get("publisherNodeR")),
                    Web3.to_bytes(hexstr=request.get("publisherNodeVS")),
                    request.get("identityIds"),
                    [Web3.to_bytes(hexstr=x) for x in request.get("r")],
                    [Web3.to_bytes(hexstr=x) for x in request.get("vs")],
                )
            else:
                receipt = await self._mint_knowledge_collection(
                    paranet_ka_contract,
                    paranet_token_id,
                    list(request.values()),
                )

            event_data = self.manager.blockchain_provider.decode_logs_event(
                receipt=receipt,
                contract_name="KnowledgeCollectionStorage",
                event_name="KnowledgeCollectionCreated",
            )
            collection_id = (
                int(getattr(event_data[0].get("args", {}), "id", None))
                if event_data
                else None
            )

            return KnowledgeCollectionResult(
                knowledge_collection_id=collection_id, receipt=receipt
            )

        except Exception as e:
            if allowance_increased:
                await self.decrease_knowledge_collection_allowance(allowance_gap)
            raise e

    # TODO: change self._owner to v8 compatible function
    async def get_owner(self, ual: UAL) -> Address:
        token_id = parse_ual(ual)["token_id"]

        return await self._owner(token_id)

    async def get_asset_storage_address(self, asset_storage_name: str) -> Address:
        return await self._get_asset_storage_address(asset_storage_name)

    async def key_is_operational_wallet(
        self, identity_id: int, key: Address, purpose: int
    ) -> bool:
        return await self._key_is_operational_wallet(identity_id, key, purpose)

    async def time_until_next_epoch(self) -> int:
        return await self._time_until_next_epoch()

    async def epoch_length(self) -> int:
        return await self._epoch_length()

    async def get_stake_weighted_average_ask(self) -> int:
        return await self._get_stake_weighted_average_ask()

    async def get_block(self, block_identifier: str | int):
        return await self._get_block(block_identifier)
